//
//  getver.c
//        get file version
//        zlel - last edit: 2011/08/03 21:47:51
//
//  compile with
//        mingw32-gcc -o getver.exe getver.c -lversion

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <winbase.h>
#include <fcntl.h>
#include "sha1.h"
#include <inttypes.h>

typedef struct t_getverinfo {
    int         valid;
    int         ver[4];
    char        creationtime[27];
    char        modifytime[27];
    char        sha1[42];
    char        filename[MAX_PATH];
    FILETIME    ftCreationTime;
    FILETIME    ftLastWriteTime;
    DWORD       dwFileAttributes;
    unsigned long long int filesize;
} getverinfo;


getverinfo getver (char *filename, int includecreationtime, int excludesha1, int nofiles, int tabmode)
{
    DWORD dwArg;
    VS_FIXEDFILEINFO *vInfo;
    VS_FIXEDFILEINFO empty;
    UINT uInfoSize;

    GET_FILEEX_INFO_LEVELS fInfoLevelId = GetFileExInfoStandard;
    LPWIN32_FILE_ATTRIBUTE_DATA lpFileInformation;
    LPSYSTEMTIME lpSystemTimeModify;
    LPSYSTEMTIME lpSystemTimeCreate;

    char *fullfilename;
    int filenamelength;

    int i;
    char *m;

    getverinfo fileinfo;

    int index;
    index=1;

    int res = 1;

    // Resolve filename
    // -----------------------------------------------------------
    filenamelength = GetFullPathName(filename, 0, NULL, NULL);
    fullfilename = (char *) malloc((filenamelength+1) * sizeof(char));
    GetFullPathName(filename, filenamelength, fullfilename, NULL);

    sprintf(fileinfo.filename, "%s", filename);
    while(m=_mbschr(fileinfo.filename, '\\')) {
        m[0] = '/'; 
    }


    // Get version information
    // -----------------------------------------------------------
    lpFileInformation = malloc(sizeof(WIN32_FILE_ATTRIBUTE_DATA));


    res = res && GetFileAttributesEx(fullfilename, fInfoLevelId, lpFileInformation);

    fileinfo.filesize = 0;
    fileinfo.valid = 0;
    fileinfo.dwFileAttributes = 0;

    if (!res) {
        free(fullfilename);
        free(lpFileInformation);
        return fileinfo;
    }

    if (lpFileInformation->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
        fileinfo.dwFileAttributes = lpFileInformation->dwFileAttributes;
        free(fullfilename);
        free(lpFileInformation);
        return fileinfo;
    }

    lpSystemTimeModify = malloc(sizeof(SYSTEMTIME));
    lpSystemTimeCreate = malloc(sizeof(SYSTEMTIME));

    if (res) {

        FILETIME filemodificationtime;
        FILETIME filecreationtime;
        FileTimeToLocalFileTime(&(lpFileInformation->ftLastWriteTime), &filemodificationtime);
        FileTimeToLocalFileTime(&(lpFileInformation->ftCreationTime), &filecreationtime);
        FileTimeToSystemTime(&filemodificationtime, lpSystemTimeModify);
        FileTimeToSystemTime(&filecreationtime, lpSystemTimeCreate);

        fileinfo.ftCreationTime = lpFileInformation->ftCreationTime;
        fileinfo.ftLastWriteTime  = lpFileInformation->ftLastWriteTime;

        DWORD dwInfoSize = GetFileVersionInfoSize(fullfilename, &dwArg);
        BYTE* lpBuff = malloc(sizeof(BYTE) * dwInfoSize);

        GetFileVersionInfo(fullfilename, 0, dwInfoSize, lpBuff);

        if (0==dwInfoSize) {
            // No version information found
            empty.dwFileVersionMS=0;
            empty.dwFileVersionLS=0;
            vInfo = &empty;
        } else {
            if(0 == VerQueryValue(lpBuff, TEXT("\\"), (LPVOID*)&vInfo, &uInfoSize))
            {
                // No version information found
                empty.dwFileVersionMS=0;
                empty.dwFileVersionLS=0;
                vInfo = &empty;
            } else {
                if(0 == uInfoSize)
                {
                    // No version information found
                    empty.dwFileVersionMS=0;
                    empty.dwFileVersionLS=0;
                    vInfo = &empty;
                }
            }
        }
        free(lpBuff);
    }


    uint8_t digest[SHA1HashSize];
    if (!excludesha1) {
        // Calculate SHA1
        // -----------------------------------------------------------
        if (res) {
            SHA1Context sha;
            int fh;
            uint8_t buffer[1024];
            int readbytes;

            SHA1Reset(&sha);
            fh = open(fullfilename, O_BINARY|O_RDONLY);
            while(readbytes=read(fh, buffer, 1024)) {
                SHA1Input(&sha, buffer, readbytes);
            }
            SHA1Result(&sha, digest); 
            close(fh);
        }
    }
    
    // Get Timezone
    // -----------------------------------------------------------
    TIME_ZONE_INFORMATION TimeZoneInformation;
    GetTimeZoneInformation(&TimeZoneInformation);

    // Output result
    // -----------------------------------------------------------
    if (res) {

        char *p;
        LARGE_INTEGER filesize;

        fileinfo.valid = 1;
        filesize.LowPart = lpFileInformation->nFileSizeLow;
        filesize.HighPart = lpFileInformation->nFileSizeHigh;
        fileinfo.filesize = filesize.QuadPart;
        fileinfo.ver[0] = HIWORD(vInfo->dwFileVersionMS);
        fileinfo.ver[1] = LOWORD(vInfo->dwFileVersionMS);
        fileinfo.ver[2] = HIWORD(vInfo->dwFileVersionLS);
        fileinfo.ver[3] = LOWORD(vInfo->dwFileVersionLS);
        sprintf(fileinfo.modifytime, "%04d-%02d-%02d %02d:%02d:%02d@%+02d:%02d ", lpSystemTimeModify->wYear, lpSystemTimeModify->wMonth, lpSystemTimeModify->wDay, lpSystemTimeModify->wHour, lpSystemTimeModify->wMinute, lpSystemTimeModify->wSecond, -(TimeZoneInformation.Bias/60), (TimeZoneInformation.Bias%60));
        if (includecreationtime==1) {
            sprintf(fileinfo.creationtime, "%04d-%02d-%02d %02d:%02d:%02d@%+02d:%02d ", lpSystemTimeCreate->wYear, lpSystemTimeCreate->wMonth, lpSystemTimeCreate->wDay, lpSystemTimeCreate->wHour, lpSystemTimeCreate->wMinute, lpSystemTimeCreate->wSecond, -(TimeZoneInformation.Bias/60),  (TimeZoneInformation.Bias%60));
        } else {
            fileinfo.creationtime[0] = (char) 0;
        }

        p = (char *) &(fileinfo.sha1[0]);
        if (!excludesha1) {
            for (i=0; i<SHA1HashSize; i++) {
                p += sprintf(p, "%0.2x", digest[i]);
            }
            p += sprintf(p, " ");
        } else {
            fileinfo.sha1[0] = (char) 0;
        }

        // printf("%d %01d,%01d,%01d,%01d %04d-%02d-%02d %02d:%02d:%02d@%+02d:%02d ", lpFileInformation->nFileSizeLow, HIWORD(vInfo->dwFileVersionMS), LOWORD(vInfo->dwFileVersionMS), HIWORD(vInfo->dwFileVersionLS), LOWORD(vInfo->dwFileVersionLS), lpSystemTimeModify->wYear, lpSystemTimeModify->wMonth, lpSystemTimeModify->wDay, lpSystemTimeModify->wHour, lpSystemTimeModify->wMinute, lpSystemTimeModify->wSecond, -(TimeZoneInformation.Bias/60), (TimeZoneInformation.Bias%60));
        // if (includecreationtime==1) {
        //     printf("%04d-%02d-%02d %02d:%02d:%02d@%+02d:%02d ", lpSystemTimeCreate->wYear, lpSystemTimeCreate->wMonth, lpSystemTimeCreate->wDay, lpSystemTimeCreate->wHour, lpSystemTimeCreate->wMinute, lpSystemTimeCreate->wSecond, -(TimeZoneInformation.Bias/60), (TimeZoneInformation.Bias%60));
        // }
        // if (!excludesha1) {
        //     for (i=0; i<SHA1HashSize; i++) {
        //         printf("%0.2x", digest[i]);
        //         p += sprintf(p, "%0.2x", digest[i]);
        //     }
        //     printf(" ");
        // }
        // printf("*%s\n", filename);

        if (!nofiles) {
            if (tabmode) {
                printf("%I64u\t%01d,%01d,%01d,%01d\t%s\t%s\t%s\t%s\n", fileinfo.filesize, fileinfo.ver[0], fileinfo.ver[1], fileinfo.ver[2],fileinfo.ver[3], fileinfo.modifytime, fileinfo.creationtime, fileinfo.sha1, fileinfo.filename);
            } else {
                printf("%I64u %01d,%01d,%01d,%01d %s%s%s*%s\n", fileinfo.filesize, fileinfo.ver[0], fileinfo.ver[1], fileinfo.ver[2],fileinfo.ver[3], fileinfo.modifytime, fileinfo.creationtime, fileinfo.sha1, fileinfo.filename);
            }
        }
    }

    free(fullfilename);
    free(lpFileInformation);
    free(lpSystemTimeModify);
    free(lpSystemTimeCreate);

    return fileinfo;

}

getverinfo getverdir (char *entryname, int briefmode, int nofiles, int directories, int recursive, int includecreationtime, int excludesha1, int tabmode, int listmode, int level)
{
        DWORD dwArg;
        GET_FILEEX_INFO_LEVELS fInfoLevelId = GetFileExInfoStandard;
        UINT uInfoSize;
        VS_FIXEDFILEINFO *vInfo;
        VS_FIXEDFILEINFO empty;
        getverinfo childgetverinfo;
        getverinfo tmpgetverinfo;
        int filenamelength;
        TCHAR szDir[MAX_PATH];
        WIN32_FIND_DATA ffd;
        LARGE_INTEGER filesize;
        size_t length_of_arg;
        HANDLE hFind = INVALID_HANDLE_VALUE;
        DWORD dwError=0;
        SYSTEMTIME SystemTimeModify;
        SYSTEMTIME SystemTimeCreate;
        int i;
        char *m;

        childgetverinfo.valid = 0;
        childgetverinfo.filesize = 0;

        sprintf(childgetverinfo.filename, "%s", entryname);
        while(m=_mbschr(childgetverinfo.filename, '\\')) {
            m[0] = '/';
        }
        childgetverinfo.modifytime[0] = (char) 0;
        childgetverinfo.creationtime[0] = (char) 0;
        childgetverinfo.sha1[0] = (char) 0;

        strncpy(szDir, entryname, MAX_PATH);

        tmpgetverinfo = getver(entryname, includecreationtime, excludesha1, 1, tabmode);

        if (recursive && (tmpgetverinfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
            
            while(entryname[_mbslen(entryname)-1]=='/') {
                entryname[_mbslen(entryname)-1]=(char)0;
            }

            while(childgetverinfo.filename[_mbslen(childgetverinfo.filename)-1]=='/') {
                childgetverinfo.filename[_mbslen(childgetverinfo.filename)-1]=(char)0;
            }
            _mbscat(childgetverinfo.filename, "/");

            sprintf(szDir, "%s\\*", entryname);

            // Find the first file in the directory.
            hFind = FindFirstFile(szDir, &ffd);

            if (INVALID_HANDLE_VALUE != hFind) {
                do {
                    sprintf(szDir, "%s\\%s", entryname, ffd.cFileName);
                    if ((ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) && (strcmp(".", ffd.cFileName)!=0) && (strcmp("..", ffd.cFileName)!=0)) {


                        if ((childgetverinfo.valid==0)||(CompareFileTime(&(childgetverinfo.ftCreationTime), &(ffd.ftCreationTime)) > 0 )) {
                             childgetverinfo.ftCreationTime = ffd.ftCreationTime;
                        }

                        tmpgetverinfo = getverdir(szDir, briefmode, nofiles, directories, recursive, includecreationtime, excludesha1, tabmode, listmode, level+1);

                    } else if ((strcmp(".", ffd.cFileName)!=0) && (strcmp("..", ffd.cFileName)!=0)) {
                        
                        if (!briefmode) {
                            tmpgetverinfo = getver(szDir, includecreationtime, excludesha1, nofiles, tabmode);
                        } else {
                            filesize.LowPart = ffd.nFileSizeLow;
                            filesize.HighPart = ffd.nFileSizeHigh;
                            tmpgetverinfo.filesize = filesize.QuadPart;
                            tmpgetverinfo.valid = 1;
                            tmpgetverinfo.ftCreationTime = ffd.ftCreationTime;
                            tmpgetverinfo.ftLastWriteTime = ffd.ftLastWriteTime;
                        }

                    }
                    if (tmpgetverinfo.valid!=0) {
                         if ((childgetverinfo.valid==0)||(CompareFileTime(&(childgetverinfo.ftCreationTime), &(tmpgetverinfo.ftCreationTime)) > 0 )) {
                             childgetverinfo.ftCreationTime = tmpgetverinfo.ftCreationTime;
                             sprintf(childgetverinfo.creationtime, "%s", tmpgetverinfo.creationtime);
                         }
                         if ((childgetverinfo.valid==0)||(CompareFileTime(&(childgetverinfo.ftLastWriteTime), &(tmpgetverinfo.ftLastWriteTime)) < 0 )) {
                             childgetverinfo.ftLastWriteTime = tmpgetverinfo.ftLastWriteTime;
                             sprintf(childgetverinfo.modifytime, "%s", tmpgetverinfo.modifytime);
                         }
                         if (
                            (childgetverinfo.valid==0) || 
                            (childgetverinfo.ver[0] < tmpgetverinfo.ver[0]) || 
                            ((childgetverinfo.ver[0] == tmpgetverinfo.ver[0]) && (childgetverinfo.ver[1] < tmpgetverinfo.ver[1])) || 
                            ((childgetverinfo.ver[0] == tmpgetverinfo.ver[0]) && (childgetverinfo.ver[1] == tmpgetverinfo.ver[1]) && (childgetverinfo.ver[2] < tmpgetverinfo.ver[2])) || 
                            ((childgetverinfo.ver[0] == tmpgetverinfo.ver[0]) && (childgetverinfo.ver[1] == tmpgetverinfo.ver[1]) && (childgetverinfo.ver[2] == tmpgetverinfo.ver[2]) && (childgetverinfo.ver[3] < tmpgetverinfo.ver[3])) 
                         ) {
                            childgetverinfo.ver[0] = tmpgetverinfo.ver[0];
                            childgetverinfo.ver[1] = tmpgetverinfo.ver[1];
                            childgetverinfo.ver[2] = tmpgetverinfo.ver[2];
                            childgetverinfo.ver[3] = tmpgetverinfo.ver[3];
                         } 
                         childgetverinfo.sha1[0] = (char) 0;
                         if (childgetverinfo.valid==0) {
                             childgetverinfo.filesize = 0;
                             childgetverinfo.valid = 1;
                         }
                         childgetverinfo.valid = 1;
                         childgetverinfo.filesize += tmpgetverinfo.filesize;
                    }       
                } while (FindNextFile(hFind, &ffd) != 0);

                FILETIME filemodificationtime;
                FILETIME filecreationtime;
                FileTimeToLocalFileTime(&(childgetverinfo.ftLastWriteTime), &filemodificationtime);
                FileTimeToLocalFileTime(&(childgetverinfo.ftCreationTime), &filecreationtime);
                FileTimeToSystemTime(&filemodificationtime, &SystemTimeModify);
                FileTimeToSystemTime(&filecreationtime, &SystemTimeCreate);
                if (briefmode) {
                    childgetverinfo.ver[0] = 0;
                    childgetverinfo.ver[1] = 0;
                    childgetverinfo.ver[2] = 0;
                    childgetverinfo.ver[3] = 0;
                    TIME_ZONE_INFORMATION TimeZoneInformation;
                    GetTimeZoneInformation(&TimeZoneInformation);
                    sprintf(childgetverinfo.modifytime, "%04d-%02d-%02d %02d:%02d:%02d@%+02d:%02d ", SystemTimeModify.wYear, SystemTimeModify.wMonth, SystemTimeModify.wDay, SystemTimeModify.wHour, SystemTimeModify.wMinute, SystemTimeModify.wSecond, -(TimeZoneInformation.Bias/60), (TimeZoneInformation.Bias%60));
                    sprintf(childgetverinfo.creationtime, "%04d-%02d-%02d %02d:%02d:%02d@%+02d:%02d ", SystemTimeCreate.wYear, SystemTimeCreate.wMonth, SystemTimeCreate.wDay, SystemTimeCreate.wHour, SystemTimeCreate.wMinute, SystemTimeCreate.wSecond, -(TimeZoneInformation.Bias/60), (TimeZoneInformation.Bias%60));
                }
               
                if (directories || (level==0)) {
                    if (childgetverinfo.valid==1) {
                        if (tabmode) {
                            printf("%I64u\t%01d,%01d,%01d,%01d\t%s\t%s\t%s\t%s\n", childgetverinfo.filesize, childgetverinfo.ver[0], childgetverinfo.ver[1], childgetverinfo.ver[2],childgetverinfo.ver[3], childgetverinfo.modifytime, childgetverinfo.creationtime, childgetverinfo.sha1, childgetverinfo.filename);
                        } else {
                            printf("%I64u %01d,%01d,%01d,%01d %s%s%s*%s\n", childgetverinfo.filesize, childgetverinfo.ver[0], childgetverinfo.ver[1], childgetverinfo.ver[2], childgetverinfo.ver[3], childgetverinfo.modifytime, childgetverinfo.creationtime, childgetverinfo.sha1, childgetverinfo.filename);
                        }
                    } else {
                        if (tabmode) {
                            printf("%d\t%01d,%01d,%01d,%01d\t%s\t%s\t%s\t%s\n", 0, childgetverinfo.ver[0], childgetverinfo.ver[1], childgetverinfo.ver[2],childgetverinfo.ver[3], childgetverinfo.modifytime, childgetverinfo.creationtime, childgetverinfo.sha1, childgetverinfo.filename);
                        } else {
                            printf("%d %01d,%01d,%01d,%01d %s%s%s*%s\n", 0, childgetverinfo.ver[0], childgetverinfo.ver[1], childgetverinfo.ver[2], childgetverinfo.ver[3], childgetverinfo.modifytime, childgetverinfo.creationtime, childgetverinfo.sha1, childgetverinfo.filename);
                        }
                    }
                }
            }
            FindClose(hFind);
        } else if(!(tmpgetverinfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
            childgetverinfo = getver(entryname, includecreationtime, excludesha1, nofiles, tabmode);
        }

        return childgetverinfo;


}

int main (int argc, char *argv[ ], char *envp[ ])
{
    int i;
    char *p;
    int start = 1;
    int listmode = 0;
    int tabmode = 0;
    int excludesha1 = 0;
    int includecreationtime = 0;
    int nofiles = 0;
    int directories = 0;
    int recursive = 0;
    int briefmode = 0;

    while ( (argc==1) || (argv[start][0]=='-')) {
        if ( ((argc-start+1)<=2) || (strcmp(argv[start],"--help")==0)|| (strcmp(argv[start],"-h")==0)|| (strcmp(argv[start],"-v")==0) || (strcmp(argv[start],"--version")==0)) {
            p=argv[0] + _mbslen(argv[0]);
            while ((*p!='\\') && (*p!='/') && (p>argv[0])) { 
                p--;
            }
            if ((p[0]=='\\') ||  (p[0]=='/')) {
                p++;
            }
           printf("getver: %s [-c|--creationdate] [-d|--directories] [-F|--nofiles] [-r|--recursive] [--excludesha1|-S] [--brief] <filename> [ <filename> ... ]\n", p);
           printf("display file version information, optionally summarizing files as directory entries in the format\n");
           printf("<size> <version> <modifydate> <modifytime> <sha1> *<filename>\n");
           return;
        } else if ( (strcmp(argv[start],"--brief")==0) || (strcmp(argv[start],"-b")==0)) {
            briefmode = 1;
            start++; 
        } else if ( (strcmp(argv[start],"--tabs")==0) || (strcmp(argv[start],"-t")==0)) {
            tabmode = 1;
            start++; 
        } else if ( (strcmp(argv[start],"--recursive")==0) || (strcmp(argv[start],"-r")==0)) {
            recursive = 1;
            start++; 
        } else if ( (strcmp(argv[start],"--nofiles")==0) || (strcmp(argv[start],"-F")==0)) {
            nofiles = 1;
            start++; 
        } else if ( (strcmp(argv[start],"--directories")==0) || (strcmp(argv[start],"-d")==0)) {
            directories = 1;
            start++; 
        } else if ( (strcmp(argv[start],"--excludesha1")==0) || (strcmp(argv[start],"-S")==0)) {
            excludesha1 = 1;
            start++; 
        } else if ( (strcmp(argv[start],"--creationdate")==0) || (strcmp(argv[start],"-c")==0)) {
            includecreationtime = 1;
            start++; 
        } else {
            p=argv[0] + _mbslen(argv[0]);
            while ((*p!='\\') && (*p!='/') && (p>argv[0])) { 
                p--;
            }
            if ((p[0]=='\\') ||  (p[0]=='/')) {
                p++;
            }
            printf("%s: invalid option %s", p, argv[start]);
            return;
        }
    }
    for (i=start; i<argc; i++) {
        char *sp;
        sp = &argv[i][0];
        while (_mbslen(sp) && (sp[0]=='"')) { sp++; };
        while (_mbslen(argv[i]) && (argv[i][_mbslen(argv[i])-1]=='"')) { argv[i][_mbslen(argv[i])-1]='\0'; }
        getverdir(sp, briefmode, nofiles, directories, recursive, includecreationtime, excludesha1, tabmode, listmode, 0); 
    }
}


